"""The scheduler to run jobs on Slurm"""
import asyncio
from pathlib import Path
from typing import Type

from ..defaults import JobStatus
from ..job import Job
from ..scheduler import Scheduler
from ..utils import a_read_text


class SlurmJob(Job):
    """Slurm job"""

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self._srun_opts = None

    # @property
    # def strcmd(self) -> str:
    #     """Get the string representation of the command"""
    #     cmd = super().strcmd
    #     if self._srun_opts is None:
    #         return cmd

    #     srun_opts = self._srun_opts.copy()
    #     cmd_parts = [srun_opts.pop('srun', 'srun')]
    #     for key, val in self._srun_opts.items():
    #         if len(key) == 1:
    #             fmt = f'-{key} {val}'
    #         else:
    #             fmt = f'--{key}={val}'
    #         cmd_parts.append(fmt.format(key=key, val=val))
    #     cmd_parts.append(cmd)
    #     return ' '.join(cmd_parts)
    # def wrap_cmd(self, scheduler: Scheduler, srun: str) -> str:
    def wrap_cmd(self, scheduler: Scheduler) -> str:
        """Wrap the command to enable status, returncode, cleaning when
        job exits

        Args:
            scheduler: The scheduler

        Returns:
            The wrapped script
        """
        options = {
            key[6:]: val
            for key, val in scheduler.config.items()
            if key.startswith("slurm_")
        }
        sbatch_options = {
            key[7:]: val
            for key, val in scheduler.config.items()
            if key.startswith("sbatch_")
        }
        options.update(sbatch_options)
        # use_srun = options.pop('use_srun', False)
        # if not use_srun:
        #     self._srun_opts = None
        # else:
        #     self._srun_opts = {
        #         key[5:]: val
        #         for key, val in scheduler.config.items()
        #         if key.startswith('srun_')
        #     }
        #     self._srun_opts['srun'] = srun

        jobname_prefix = scheduler.config.get(
            "scheduler_jobprefix",
            scheduler.name,
        )
        options["job-name"] = f"{jobname_prefix}.{self.index}"
        options["chdir"] = str(Path.cwd().resolve())
        options["output"] = self.stdout_file
        options["error"] = self.stderr_file

        options_list = []
        for key, val in options.items():
            key = key.replace("_", "-")
            if len(key) == 1:
                fmt = "#SBATCH -{key} {val}"
            else:
                fmt = "#SBATCH --{key}={val}"
            options_list.append(fmt.format(key=key, val=val))

        options_str = "\n".join(options_list)

        return self.CMD_WRAPPER_TEMPLATE.format(
            shebang=f"#!{self.CMD_WRAPPER_SHELL}\n{options_str}\n",
            prescript=scheduler.config.prescript,
            postscript=scheduler.config.postscript,
            job=self,
            status=JobStatus,
        )


class SlurmScheduler(Scheduler):
    """The Slurm scheduler

    Attributes:
        name: The name of the scheduler
        job_class: The job class

    Args:
        sbatch: path to sbatch command
        squeue: path to squeue command
        scancel: path to scancel command
        slurm_*: Slurm options for sbatch.
        ... other Scheduler args
    """

    name: str = "slurm"
    job_class: Type[Job] = SlurmJob

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sbatch = self.config.get("sbatch", "sbatch")
        # self.srun = self.config.get('srun', 'srun')
        self.scancel = self.config.get("scancel", "scancel")
        self.squeue = self.config.get("squeue", "squeue")

    async def submit_job(self, job: Job) -> str:
        """Submit a job to Slurm

        Args:
            job: The job

        Returns:
            The job id
        """
        proc = await asyncio.create_subprocess_exec(
            self.sbatch,
            # str(await job.wrapped_script(self. self.srun)),
            str(await job.wrapped_script(self)),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        rc = await proc.wait()
        if rc != 0:
            stderr = await proc.stderr.read()
            raise Exception(stderr.decode().strip())

        stdout = await proc.stdout.read()
        # salloc: Granted job allocation 65537
        # sbatch: Submitted batch job 65537
        return stdout.decode().strip().split()[-1]

    async def kill_job(self, job: Job):
        """Kill a job on Slurm

        Args:
            job: The job
        """
        proc = await asyncio.create_subprocess_exec(
            self.scancel,
            str(job.jid),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await proc.wait()

    async def job_is_running(self, job: Job) -> bool:
        """Tell if a job is really running, not only the job.jid_file

        In case where the jid file is not cleaned when job is done.

        Args:
            job: The job

        Returns:
            True if it is, otherwise False
        """
        try:
            jid = await a_read_text(job.jid_file)
        except FileNotFoundError:
            return False

        if not jid:
            return False

        proc = await asyncio.create_subprocess_exec(
            self.squeue,
            "-j",
            jid,
            "--noheader",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await proc.wait()
        if proc.returncode != 0:
            return False

        # ['8792', 'queue', 'merge', 'user', 'R', '7:34:34', '1', 'server']
        st = (
            await proc.stdout.read()  # type: ignore
        ).decode().strip().split()[4]
        # If job is still take resources, it is running
        return st in (
            "R",
            "RUNNING",
            "PD",
            "PENDING",
            "CG",
            "COMPLETING",
            "S",
            "SUSPENDED",
            "CF",
            "CONFIGURING",
            # Job is being held after requested reservation was deleted.
            "RD",
            "RESV_DEL_HOLD",
            # Job is being requeued by a federation.
            "RF",
            "REQUEUE_FED",
            # Held job is being requeued.
            "RH",
            "REQUEUE_HOLD",
            # Completing job is being requeued.
            "RQ",
            "REQUEUED",
            # Job is about to change size.
            "RS",
            "RESIZING",
            # Sibling was removed from cluster due to other cluster
            # starting the job.
            "RV",
            "REVOKED",
            # The job was requeued in a special state. This state can be set by
            # users, typically in EpilogSlurmctld, if the job has terminated
            # with a particular exit value.
            "SE",
            "SPECIAL_EXIT",
            # Job is staging out files.
            "SO",
            "STAGE_OUT",
            # Job has an allocation, but execution has been stopped with
            # SIGSTOP signal. CPUS have been retained by this job.
            "ST",
            "STOPPED",
        )



"""The xqute module"""
from __future__ import annotations

import asyncio
import functools
import signal
from collections import deque
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Type

from .defaults import (
    DEFAULT_JOB_METADIR,
    DEFAULT_JOB_ERROR_STRATEGY,
    DEFAULT_JOB_NUM_RETRIES,
    DEFAULT_SCHEDULER_FORKS,
    DEFAULT_JOB_SUBMISSION_BATCH,
    JobErrorStrategy,
    JobStatus,
)
from .utils import logger
from .plugin import plugin
from .schedulers import get_scheduler

if TYPE_CHECKING:  # pragma: no cover
    from .scheduler import Scheduler
    from .job import Job


class Xqute:
    """The main class of the package

    Attributes:
        name: The name, used in logger
        EMPTY_BUFFER_SLEEP_TIME: The time to sleep while waiting when
            the buffer is empty

        jobs: The jobs registry
        plugins: The plugins to be enabled or disabled
            to disable a plugin, using `no:plugin_name`
            either all plugin names should be prefixed with 'no:' or none
            of them should
        job_submission_batch: The number of consumers to submit jobs

        _job_metadir: The job meta directory
        _job_error_strategy: The strategy when there is error happened
        _job_num_retries: The number of retries when strategy is retry

        _cancelling: A mark to mark whether a shutting down event
            is triggered (True for natual cancelling, the signale for
            cancelling with a signal, SIGINT for example)

        buffer_queue: A buffer queue to save the pushed jobs
        queue: The job queue
        scheduler: The scheduler
        task: The task of producer and consumers

    Args:
        scheduler: The scheduler class or name
        plugins: The plugins to be enabled or disabled
            to disable a plugin, using `no:plugin_name`
            either all plugin names should be prefixed with 'no:' or none
            of them should
            To enabled plugins, objects are
        job_metadir: The job meta directory
        job_submission_batch: The number of consumers to submit jobs
        job_error_strategy: The strategy when there is error happened
        job_num_retries: Max number of retries when job_error_strategy is retry
        scheduler_forks: Max number of job forks
        **scheduler_opts: Additional keyword arguments for scheduler
    """

    name: str = "Xqute"
    EMPTY_BUFFER_SLEEP_TIME: int = 1

    def __init__(
        self,
        scheduler: str | Type[Scheduler] = "local",
        plugins: List[Any] | None = None,
        *,
        job_metadir: PathLike = DEFAULT_JOB_METADIR,
        job_submission_batch: int = DEFAULT_JOB_SUBMISSION_BATCH,
        job_error_strategy: str = DEFAULT_JOB_ERROR_STRATEGY,
        job_num_retries: int = DEFAULT_JOB_NUM_RETRIES,
        scheduler_forks: int = DEFAULT_SCHEDULER_FORKS,
        scheduler_prescript: str = "",
        scheduler_postscript: str = "",
        **scheduler_opts,
    ) -> None:
        """Construct"""
        self.jobs: List[Job] = []

        if plugins is not None:
            no_plugins = [
                isinstance(plug, str) and plug.startswith("no:")
                for plug in plugins
            ]
            if any(no_plugins) and not all(no_plugins):
                raise ValueError(
                    'Either all plugin names start with "no:" or '
                    "none of them does."
                )
            if all(no_plugins):
                self.plugin_context = plugin.plugins_but_context(
                    plug[3:] for plug in plugins
                )
            else:
                self.plugin_context = plugin.plugins_only_context(plugins)
        else:
            self.plugin_context = plugin.plugins_only_context(plugins)

        if self.plugin_context:
            self.plugin_context.__enter__()

        logger.info(
            "/%s Enabled plugins: %s",
            self.name,
            plugin.get_enabled_plugin_names(),
        )

        self.job_submission_batch = job_submission_batch
        self._job_metadir = job_metadir
        Path(self._job_metadir).mkdir(exist_ok=True)
        self._job_error_strategy = job_error_strategy
        self._job_num_retries = job_num_retries

        self._cancelling: bool | signal.Signals = False

        self.buffer_queue: deque = deque()
        self.queue: asyncio.Queue = asyncio.Queue()
        self.scheduler = get_scheduler(scheduler)(
            scheduler_forks,
            scheduler_prescript,
            scheduler_postscript,
            **scheduler_opts,
        )

        # requires to be defined in a loop
        loop = asyncio.get_running_loop()
        for sig in (signal.SIGTERM, signal.SIGINT):
            loop.add_signal_handler(sig, functools.partial(self.cancel, sig))

        self.task = asyncio.gather(
            self._producer(),
            *(self._consumer(i) for i in range(self.job_submission_batch)),
        )
        plugin.hooks.on_init(self)

    def cancel(self, sig: signal.Signals | None = None) -> None:
        """Cancel the producer-consumer task

        `self._cancelling` will be set to `signaled` if sig is provided,
        otherwise it will be set to `True`

        Args:
            sig: Whether this cancelling is caused by a signal
        """
        self._cancelling = True
        if sig:
            self._cancelling = sig
            logger.warning(
                "/%s Got signal %r, trying a graceful " "shutdown ...",
                self.name,
                sig.name,
            )

        if plugin.hooks.on_shutdown(self, sig) is not False:
            self.task.cancel()

    async def _producer(self):
        """The producer"""

        while True:
            if not self.buffer_queue:
                logger.debug(
                    "/%s Buffer queue is empty, waiting ...", self.name
                )
                await asyncio.sleep(self.EMPTY_BUFFER_SLEEP_TIME)
                continue

            job = self.buffer_queue.popleft()
            if not await self.scheduler.polling_jobs(
                self.jobs,
                "can_submit",
                self._job_error_strategy == JobErrorStrategy.HALT,
            ):
                logger.debug("/%s Hit max forks of scheduler ...", self.name)
                await asyncio.sleep(0.1)
                self.buffer_queue.appendleft(job)
                continue

            job.status = JobStatus.QUEUED
            await self.queue.put(job)

    async def _consumer(self, index: int) -> None:
        """The consumer

        Args:
            index: The index of the consumer
        """
        while True:
            job = await self.queue.get()
            logger.debug(
                "/%s 'Consumer-%s' submitting %s", self.name, index, job
            )
            await self.scheduler.submit_job_and_update_status(job)
            self.queue.task_done()

    async def put(self, cmd: Job | str | List[str]) -> None:
        """Put a command into the buffer

        Args:
            cmd: The command
        """
        if isinstance(cmd, self.scheduler.job_class):
            job = cmd
            if job._error_retry is None:
                job._error_retry = (
                    self._job_error_strategy == JobErrorStrategy.RETRY
                )
            if job._num_retries is None:
                job._num_retries = self._job_num_retries
        else:
            job = self.scheduler.job_class(
                len(self.jobs),
                cmd,  # type: ignore
                self._job_metadir,
                self._job_error_strategy == JobErrorStrategy.RETRY,
                self._job_num_retries,
            )
        await plugin.hooks.on_job_init(self.scheduler, job)
        self.jobs.append(job)
        logger.info("/%s Pushing job: %r", self.name, job)

        self.buffer_queue.append(job)
        await plugin.hooks.on_job_queued(self.scheduler, job)

    async def _polling_jobs(self) -> None:
        """Polling the jobs to see if they are all done.

        If yes, cancel the producer-consumer task natually.
        """
        while (
            self._cancelling is False
            and not await self.scheduler.polling_jobs(
                self.jobs,
                "all_done",
                self._job_error_strategy == JobErrorStrategy.HALT,
            )
        ):
            await asyncio.sleep(1.0)

        if self._cancelling is False:
            self.cancel()

    async def _await_task(self) -> None:
        """Await the producer-consumer task and catch the CancelledError"""
        try:
            await self.task
        except asyncio.CancelledError:
            logger.debug("/%s Stoping producer and consumer ...", self.name)
            if self._cancelling not in (True, False):  # signaled
                await self.scheduler.kill_running_jobs(self.jobs)

        logger.info("/%s Done!", self.name)

    async def run_until_complete(self) -> None:
        """Wait until all jobs complete"""
        logger.debug(
            "/%s Done feeding jobs, waiting for jobs to be done ...", self.name
        )
        try:
            await asyncio.gather(self._polling_jobs(), self._await_task())
        finally:
            if self.plugin_context:
                self.plugin_context.__exit__()



"""The scheduler to run jobs locally"""
import asyncio
import os
from typing import Type

from ..defaults import JobStatus
from ..job import Job
from ..scheduler import Scheduler
from ..utils import a_read_text


class LocalJob(Job):
    """Local job"""

    def wrap_cmd(self, scheduler: Scheduler) -> str:
        """Wrap the command for the scheduler to submit and run

        Args:
            scheduler: The scheduler

        Returns:
            The wrapped script
        """
        return self.CMD_WRAPPER_TEMPLATE.format(
            shebang=f"#!{self.CMD_WRAPPER_SHELL}",
            prescript=scheduler.config.prescript,
            postscript=scheduler.config.postscript,
            job=self,
            status=JobStatus,
        )


class LocalScheduler(Scheduler):
    """The local scheduler

    Attributes:
        name: The name of the scheduler
        job_class: The job class
    """

    name = "local"
    job_class: Type[Job] = LocalJob

    async def submit_job(self, job: Job) -> int:
        """Submit a job locally

        Args:
            job: The job

        Returns:
            The process id
        """
        proc = await asyncio.create_subprocess_exec(
            job.CMD_WRAPPER_SHELL,
            str(await job.wrapped_script(self)),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        # don't await for the results, as this will run the real command
        return proc.pid

    async def kill_job(self, job: Job):
        """Kill a job asynchronously

        Args:
            job: The job
        """
        try:
            os.killpg(int(job.jid), 9)
        except Exception:  # pragma: no cover
            pass

    async def job_is_running(self, job: Job) -> bool:
        """Tell if a job is really running, not only the job.jid_file

        In case where the jid file is not cleaned when job is done.

        Args:
            job: The job

        Returns:
            True if it is, otherwise False
        """
        try:
            jid = int(await a_read_text(job.jid_file))
        except (ValueError, TypeError, FileNotFoundError):
            return False

        if jid <= 0:
            return False

        try:
            os.kill(jid, 0)
        except Exception:  # pragma: no cover
            return False

        return True
